// Created on: 1992-04-06
// Created by: Christian CAILLET
// Copyright (c) 1992-1999 Matra Datavision
// Copyright (c) 1999-2014 OPEN CASCADE SAS
//
// This file is part of Open CASCADE Technology software library.
//
// This library is free software; you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License version 2.1 as published
// by the Free Software Foundation, with special exception defined in the file
// OCCT_LGPL_EXCEPTION.txt. Consult the file LICENSE_LGPL_21.txt included in OCCT
// distribution for complete text of the license and disclaimer of any warranty.
//
// Alternatively, this file may be used under the terms of Open CASCADE
// commercial license or contractual agreement.

#include <IGESData_SpecificLib.hxx>

#include <IGESData_NodeOfSpecificLib.hxx>
#include <Standard_NoSuchObject.hxx>
#include <IGESData_IGESEntity.hxx>
#include <IGESData_SpecificModule.hxx>
#include <IGESData_Protocol.hxx>
#include <IGESData_GlobalNodeOfSpecificLib.hxx>
#include <Standard_Transient.hxx>

#define TheObject Handle(IGESData_IGESEntity)
#define TheObject_hxx <IGESData_IGESEntity.hxx>
#define Handle_TheModule Handle(IGESData_SpecificModule)
#define TheModule IGESData_SpecificModule
#define TheModule_hxx <IGESData_SpecificModule.hxx>
#define Handle_TheProtocol Handle(IGESData_Protocol)
#define TheProtocol IGESData_Protocol
#define TheProtocol_hxx <IGESData_Protocol.hxx>
#define LibCtl_GlobalNode IGESData_GlobalNodeOfSpecificLib
#define LibCtl_GlobalNode_hxx <IGESData_GlobalNodeOfSpecificLib.hxx>
#define LibCtl_Node IGESData_NodeOfSpecificLib
#define LibCtl_Node_hxx <IGESData_NodeOfSpecificLib.hxx>
#define Handle_LibCtl_GlobalNode Handle(IGESData_GlobalNodeOfSpecificLib)
#define Handle_LibCtl_Node Handle(IGESData_NodeOfSpecificLib)
#define LibCtl_Library IGESData_SpecificLib
#define LibCtl_Library_hxx <IGESData_SpecificLib.hxx>
#include "../LibCtl/LibCtl_Library.gxx"
