// Created on: 1996-07-03
// Created by: Joelle CHAUVET
// Copyright (c) 1996-1999 Matra Datavision
// Copyright (c) 1999-2014 OPEN CASCADE SAS
//
// This file is part of Open CASCADE Technology software library.
//
// This library is free software; you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License version 2.1 as published
// by the Free Software Foundation, with special exception defined in the file
// OCCT_LGPL_EXCEPTION.txt. Consult the file LICENSE_LGPL_21.txt included in OCCT
// distribution for complete text of the license and disclaimer of any warranty.
//
// Alternatively, this file may be used under the terms of Open CASCADE
// commercial license or contractual agreement.

#include <AdvApprox_DichoCutting.hxx>
#include <Precision.hxx>

AdvApprox_DichoCutting::AdvApprox_DichoCutting() {}

Standard_Boolean AdvApprox_DichoCutting::Value(const Standard_Real a,
                                               const Standard_Real b,
                                               Standard_Real&      cuttingvalue) const
{
  //  longueur minimum d'un intervalle pour F(U,V) : EPS1=1.e-9 (cf.MEPS1)
  constexpr Standard_Real lgmin = 10 * Precision::PConfusion();
  cuttingvalue                  = (a + b) / 2;
  return (Abs(b - a) >= 2 * lgmin);
}
