// Copyright (c) 1999-2014 OPEN CASCADE SAS
//
// This file is part of Open CASCADE Technology software library.
//
// This library is free software; you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License version 2.1 as published
// by the Free Software Foundation, with special exception defined in the file
// OCCT_LGPL_EXCEPTION.txt. Consult the file LICENSE_LGPL_21.txt included in OCCT
// distribution for complete text of the license and disclaimer of any warranty.
//
// Alternatively, this file may be used under the terms of Open CASCADE
// commercial license or contractual agreement.

#include <StepVisual_CoordinatesList.hxx>
#include <StepVisual_TessellatedItem.hxx>
IMPLEMENT_STANDARD_RTTIEXT(StepVisual_CoordinatesList, StepVisual_TessellatedItem)

StepVisual_CoordinatesList::StepVisual_CoordinatesList() {}

void StepVisual_CoordinatesList::Init(const Handle(TCollection_HAsciiString)& theName,
                                      const Handle(TColgp_HArray1OfXYZ)&      thePoints)
{
  StepRepr_RepresentationItem::Init(theName);
  myPoints = thePoints;
}

Handle(TColgp_HArray1OfXYZ) StepVisual_CoordinatesList::Points() const
{
  return myPoints;
}
