// Created on: 2002-12-10
// Created by: data exchange team
// Copyright (c) 2002-2014 OPEN CASCADE SAS
//
// This file is part of Open CASCADE Technology software library.
//
// This library is free software; you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License version 2.1 as published
// by the Free Software Foundation, with special exception defined in the file
// OCCT_LGPL_EXCEPTION.txt. Consult the file LICENSE_LGPL_21.txt included in OCCT
// distribution for complete text of the license and disclaimer of any warranty.
//
// Alternatively, this file may be used under the terms of Open CASCADE
// commercial license or contractual agreement.

// Generator:	ExpToCas (EXPRESS -> CASCADE/XSTEP Translator) V2.0

#include <StepElement_MeasureOrUnspecifiedValueMember.hxx>
#include <TCollection_HAsciiString.hxx>

IMPLEMENT_STANDARD_RTTIEXT(StepElement_MeasureOrUnspecifiedValueMember, StepData_SelectNamed)

static Standard_CString aContextDependedMeasure = "CONTEXT_DEPENDENT_MEASURE";
static Standard_CString anUnspecifiedValue      = "UNSPECIFIED_VALUE";

//=================================================================================================

StepElement_MeasureOrUnspecifiedValueMember::StepElement_MeasureOrUnspecifiedValueMember()
    : mycase(0)
{
}

//=================================================================================================

Standard_Boolean StepElement_MeasureOrUnspecifiedValueMember::HasName() const
{
  return mycase > 0;
}

//=================================================================================================

Standard_CString StepElement_MeasureOrUnspecifiedValueMember::Name() const
{
  switch (mycase)
  {
    case 1:
      return aContextDependedMeasure;
    case 2:
      return anUnspecifiedValue;
    default:
      break;
  }
  return "";
}

//=================================================================================================

static Standard_Integer CompareNames(const Standard_CString name, Standard_Integer& /*numen*/)
{
  Standard_Integer thecase = 0;
  if (!name || name[0] == '\0')
    thecase = 0;
  else if (!strcmp(name, aContextDependedMeasure))
  {
    thecase = 1;
  }
  else if (!strcmp(name, anUnspecifiedValue))
  {
    thecase = 2;
  }
  return thecase;
}

//=================================================================================================

Standard_Boolean StepElement_MeasureOrUnspecifiedValueMember::SetName(const Standard_CString name)
{
  Standard_Integer numit = 0;
  mycase                 = CompareNames(name, numit);
  if (numit)
    SetInteger(numit);
  return (mycase > 0);
}

//=================================================================================================

Standard_Boolean StepElement_MeasureOrUnspecifiedValueMember::Matches(
  const Standard_CString name) const
{
  Standard_Integer numit   = 0;
  Standard_Integer thecase = CompareNames(name, numit);
  return (mycase == thecase);
}
