// Created on: 2001-12-28
// Created by: Andrey BETENEV
// Copyright (c) 2001-2014 OPEN CASCADE SAS
//
// This file is part of Open CASCADE Technology software library.
//
// This library is free software; you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License version 2.1 as published
// by the Free Software Foundation, with special exception defined in the file
// OCCT_LGPL_EXCEPTION.txt. Consult the file LICENSE_LGPL_21.txt included in OCCT
// distribution for complete text of the license and disclaimer of any warranty.
//
// Alternatively, this file may be used under the terms of Open CASCADE
// commercial license or contractual agreement.

// Generator:	ExpToCas (EXPRESS -> CASCADE/XSTEP Translator) V1.1

#include <Interface_Check.hxx>
#include <Interface_EntityIterator.hxx>
#include "RWStepShape_RWEdgeBasedWireframeShapeRepresentation.pxx"
#include <StepData_StepReaderData.hxx>
#include <StepData_StepWriter.hxx>
#include <StepRepr_RepresentationContext.hxx>
#include <StepRepr_RepresentationItem.hxx>
#include <StepShape_EdgeBasedWireframeShapeRepresentation.hxx>

//=================================================================================================

RWStepShape_RWEdgeBasedWireframeShapeRepresentation::
  RWStepShape_RWEdgeBasedWireframeShapeRepresentation()
{
}

//=================================================================================================

void RWStepShape_RWEdgeBasedWireframeShapeRepresentation::ReadStep(
  const Handle(StepData_StepReaderData)&                         data,
  const Standard_Integer                                         num,
  Handle(Interface_Check)&                                       ach,
  const Handle(StepShape_EdgeBasedWireframeShapeRepresentation)& ent) const
{
  // Check number of parameters
  if (!data->CheckNbParams(num, 3, ach, "edge_based_wireframe_shape_representation"))
    return;

  // Inherited fields of Representation

  Handle(TCollection_HAsciiString) aRepresentation_Name;
  data->ReadString(num, 1, "representation.name", ach, aRepresentation_Name);

  Handle(StepRepr_HArray1OfRepresentationItem) aRepresentation_Items;
  Standard_Integer                             sub2 = 0;
  if (data->ReadSubList(num, 2, "representation.items", ach, sub2))
  {
    Standard_Integer num2 = sub2;
    Standard_Integer nb0  = data->NbParams(num2);
    aRepresentation_Items = new StepRepr_HArray1OfRepresentationItem(1, nb0);
    for (Standard_Integer i0 = 1; i0 <= nb0; i0++)
    {
      Handle(StepRepr_RepresentationItem) anIt0;
      data->ReadEntity(num2,
                       i0,
                       "representation.items",
                       ach,
                       STANDARD_TYPE(StepRepr_RepresentationItem),
                       anIt0);
      aRepresentation_Items->SetValue(i0, anIt0);
    }
  }

  Handle(StepRepr_RepresentationContext) aRepresentation_ContextOfItems;
  data->ReadEntity(num,
                   3,
                   "representation.context_of_items",
                   ach,
                   STANDARD_TYPE(StepRepr_RepresentationContext),
                   aRepresentation_ContextOfItems);

  // Initialize entity
  ent->Init(aRepresentation_Name, aRepresentation_Items, aRepresentation_ContextOfItems);
}

//=================================================================================================

void RWStepShape_RWEdgeBasedWireframeShapeRepresentation::WriteStep(
  StepData_StepWriter&                                           SW,
  const Handle(StepShape_EdgeBasedWireframeShapeRepresentation)& ent) const
{

  // Inherited fields of Representation

  SW.Send(ent->StepRepr_Representation::Name());

  SW.OpenSub();
  for (Standard_Integer i1 = 1; i1 <= ent->StepRepr_Representation::NbItems(); i1++)
  {
    Handle(StepRepr_RepresentationItem) Var0 = ent->StepRepr_Representation::Items()->Value(i1);
    SW.Send(Var0);
  }
  SW.CloseSub();

  SW.Send(ent->StepRepr_Representation::ContextOfItems());
}

//=================================================================================================

void RWStepShape_RWEdgeBasedWireframeShapeRepresentation::Share(
  const Handle(StepShape_EdgeBasedWireframeShapeRepresentation)& ent,
  Interface_EntityIterator&                                      iter) const
{

  // Inherited fields of Representation

  for (Standard_Integer i1 = 1; i1 <= ent->StepRepr_Representation::NbItems(); i1++)
  {
    Handle(StepRepr_RepresentationItem) Var0 = ent->StepRepr_Representation::Items()->Value(i1);
    iter.AddItem(Var0);
  }

  iter.AddItem(ent->StepRepr_Representation::ContextOfItems());
}
