// Created on: 1998-11-25
// Created by: Julia GERASIMOVA
// Copyright (c) 1998-1999 Matra Datavision
// Copyright (c) 1999-2014 OPEN CASCADE SAS
//
// This file is part of Open CASCADE Technology software library.
//
// This library is free software; you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License version 2.1 as published
// by the Free Software Foundation, with special exception defined in the file
// OCCT_LGPL_EXCEPTION.txt. Consult the file LICENSE_LGPL_21.txt included in OCCT
// distribution for complete text of the license and disclaimer of any warranty.
//
// Alternatively, this file may be used under the terms of Open CASCADE
// commercial license or contractual agreement.

#include <GeomPlate_Aij.hxx>
#include <GeomPlate_BuildAveragePlane.hxx>
#include <gp_Vec.hxx>

GeomPlate_Aij::GeomPlate_Aij()
    : Ind1(0),
      Ind2(0)
{
}

GeomPlate_Aij::GeomPlate_Aij(const Standard_Integer anInd1,
                             const Standard_Integer anInd2,
                             const gp_Vec&          aVec)
{
  Ind1 = anInd1;
  Ind2 = anInd2;
  Vec  = aVec;
}
