// Created on: 1999-11-26
// Created by: Andrey BETENEV
// Copyright (c) 1999 Matra Datavision
// Copyright (c) 1999-2014 OPEN CASCADE SAS
//
// This file is part of Open CASCADE Technology software library.
//
// This library is free software; you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License version 2.1 as published
// by the Free Software Foundation, with special exception defined in the file
// OCCT_LGPL_EXCEPTION.txt. Consult the file LICENSE_LGPL_21.txt included in OCCT
// distribution for complete text of the license and disclaimer of any warranty.
//
// Alternatively, this file may be used under the terms of Open CASCADE
// commercial license or contractual agreement.

// Generator:	ExpToCas (EXPRESS -> CASCADE/XSTEP Translator) V1.0

#include <StepBasic_Contract.hxx>
#include <StepBasic_ContractType.hxx>
#include <TCollection_HAsciiString.hxx>

IMPLEMENT_STANDARD_RTTIEXT(StepBasic_Contract, Standard_Transient)

//=================================================================================================

StepBasic_Contract::StepBasic_Contract() {}

//=================================================================================================

void StepBasic_Contract::Init(const Handle(TCollection_HAsciiString)& aName,
                              const Handle(TCollection_HAsciiString)& aPurpose,
                              const Handle(StepBasic_ContractType)&   aKind)
{

  theName = aName;

  thePurpose = aPurpose;

  theKind = aKind;
}

//=================================================================================================

Handle(TCollection_HAsciiString) StepBasic_Contract::Name() const
{
  return theName;
}

//=================================================================================================

void StepBasic_Contract::SetName(const Handle(TCollection_HAsciiString)& aName)
{
  theName = aName;
}

//=================================================================================================

Handle(TCollection_HAsciiString) StepBasic_Contract::Purpose() const
{
  return thePurpose;
}

//=================================================================================================

void StepBasic_Contract::SetPurpose(const Handle(TCollection_HAsciiString)& aPurpose)
{
  thePurpose = aPurpose;
}

//=================================================================================================

Handle(StepBasic_ContractType) StepBasic_Contract::Kind() const
{
  return theKind;
}

//=================================================================================================

void StepBasic_Contract::SetKind(const Handle(StepBasic_ContractType)& aKind)
{
  theKind = aKind;
}
