// Created on: 1992-02-03
// Created by: Christian CAILLET
// Copyright (c) 1992-1999 Matra Datavision
// Copyright (c) 1999-2014 OPEN CASCADE SAS
//
// This file is part of Open CASCADE Technology software library.
//
// This library is free software; you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License version 2.1 as published
// by the Free Software Foundation, with special exception defined in the file
// OCCT_LGPL_EXCEPTION.txt. Consult the file LICENSE_LGPL_21.txt included in OCCT
// distribution for complete text of the license and disclaimer of any warranty.
//
// Alternatively, this file may be used under the terms of Open CASCADE
// commercial license or contractual agreement.

#include <Interface_GlobalNodeOfGeneralLib.hxx>

#include <Standard_Type.hxx>

#include <Interface_GeneralModule.hxx>
#include <Interface_Protocol.hxx>
#include <Interface_GlobalNodeOfGeneralLib.hxx>
#include <Standard_Transient.hxx>
#include <Interface_GeneralLib.hxx>
#include <Interface_NodeOfGeneralLib.hxx>

#define TheObject Handle(Standard_Transient)
#define TheObject_hxx <Standard_Transient.hxx>
#define Handle_TheModule Handle(Interface_GeneralModule)
#define TheModule Interface_GeneralModule
#define TheModule_hxx <Interface_GeneralModule.hxx>
#define Handle_TheProtocol Handle(Interface_Protocol)
#define TheProtocol Interface_Protocol
#define TheProtocol_hxx <Interface_Protocol.hxx>
#define LibCtl_GlobalNode Interface_GlobalNodeOfGeneralLib
#define LibCtl_GlobalNode_hxx <Interface_GlobalNodeOfGeneralLib.hxx>
#define LibCtl_Node Interface_NodeOfGeneralLib
#define LibCtl_Node_hxx <Interface_NodeOfGeneralLib.hxx>
#define Handle_LibCtl_GlobalNode Handle(Interface_GlobalNodeOfGeneralLib)
#define Handle_LibCtl_Node Handle(Interface_NodeOfGeneralLib)
#define LibCtl_Library Interface_GeneralLib
#define LibCtl_Library_hxx <Interface_GeneralLib.hxx>
#include "../LibCtl/LibCtl_GlobalNode.gxx"
