// Copyright (c) 1999-2014 OPEN CASCADE SAS
//
// This file is part of Open CASCADE Technology software library.
//
// This library is free software; you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License version 2.1 as published
// by the Free Software Foundation, with special exception defined in the file
// OCCT_LGPL_EXCEPTION.txt. Consult the file LICENSE_LGPL_21.txt included in OCCT
// distribution for complete text of the license and disclaimer of any warranty.
//
// Alternatively, this file may be used under the terms of Open CASCADE
// commercial license or contractual agreement.

#include <StepGeom_Pcurve.hxx>
#include <StepGeom_Surface.hxx>
#include <StepRepr_DefinitionalRepresentation.hxx>
#include <TCollection_HAsciiString.hxx>

IMPLEMENT_STANDARD_RTTIEXT(StepGeom_Pcurve, StepGeom_Curve)

StepGeom_Pcurve::StepGeom_Pcurve() {}

void StepGeom_Pcurve::Init(const Handle(TCollection_HAsciiString)&            aName,
                           const Handle(StepGeom_Surface)&                    aBasisSurface,
                           const Handle(StepRepr_DefinitionalRepresentation)& aReferenceToCurve)
{
  // --- classe own fields ---
  basisSurface     = aBasisSurface;
  referenceToCurve = aReferenceToCurve;
  // --- classe inherited fields ---
  StepRepr_RepresentationItem::Init(aName);
}

void StepGeom_Pcurve::SetBasisSurface(const Handle(StepGeom_Surface)& aBasisSurface)
{
  basisSurface = aBasisSurface;
}

Handle(StepGeom_Surface) StepGeom_Pcurve::BasisSurface() const
{
  return basisSurface;
}

void StepGeom_Pcurve::SetReferenceToCurve(
  const Handle(StepRepr_DefinitionalRepresentation)& aReferenceToCurve)
{
  referenceToCurve = aReferenceToCurve;
}

Handle(StepRepr_DefinitionalRepresentation) StepGeom_Pcurve::ReferenceToCurve() const
{
  return referenceToCurve;
}
