// Created on: 2002-12-12
// Created by: data exchange team
// Copyright (c) 2002-2014 OPEN CASCADE SAS
//
// This file is part of Open CASCADE Technology software library.
//
// This library is free software; you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License version 2.1 as published
// by the Free Software Foundation, with special exception defined in the file
// OCCT_LGPL_EXCEPTION.txt. Consult the file LICENSE_LGPL_21.txt included in OCCT
// distribution for complete text of the license and disclaimer of any warranty.
//
// Alternatively, this file may be used under the terms of Open CASCADE
// commercial license or contractual agreement.

// Generator:	ExpToCas (EXPRESS -> CASCADE/XSTEP Translator) V1.2

#include <Standard_Type.hxx>
#include <StepBasic_EulerAngles.hxx>
#include <StepFEA_CurveElementInterval.hxx>
#include <StepFEA_CurveElementLocation.hxx>

IMPLEMENT_STANDARD_RTTIEXT(StepFEA_CurveElementInterval, Standard_Transient)

//=================================================================================================

StepFEA_CurveElementInterval::StepFEA_CurveElementInterval() {}

//=================================================================================================

void StepFEA_CurveElementInterval::Init(const Handle(StepFEA_CurveElementLocation)& aFinishPosition,
                                        const Handle(StepBasic_EulerAngles)&        aEuAngles)
{

  theFinishPosition = aFinishPosition;

  theEuAngles = aEuAngles;
}

//=================================================================================================

Handle(StepFEA_CurveElementLocation) StepFEA_CurveElementInterval::FinishPosition() const
{
  return theFinishPosition;
}

//=================================================================================================

void StepFEA_CurveElementInterval::SetFinishPosition(
  const Handle(StepFEA_CurveElementLocation)& aFinishPosition)
{
  theFinishPosition = aFinishPosition;
}

//=================================================================================================

Handle(StepBasic_EulerAngles) StepFEA_CurveElementInterval::EuAngles() const
{
  return theEuAngles;
}

//=================================================================================================

void StepFEA_CurveElementInterval::SetEuAngles(const Handle(StepBasic_EulerAngles)& aEuAngles)
{
  theEuAngles = aEuAngles;
}
